/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.usm.json;

import static com.openexchange.usm.json.ServletConstants.PROTOCOL;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicReference;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONException;
import org.json.JSONObject;
import com.openexchange.usm.api.configuration.ConfigurationManager;
import com.openexchange.usm.api.configuration.ConfigurationProperties;
import com.openexchange.usm.api.contenttypes.common.ContentTypeManager;
import com.openexchange.usm.api.exceptions.TemporaryDownOrBusyException;
import com.openexchange.usm.api.exceptions.USMException;
import com.openexchange.usm.api.exceptions.USMInvalidConfigurationException;
import com.openexchange.usm.api.exceptions.USMJSONException;
import com.openexchange.usm.api.exceptions.USMResourceAccessException;
import com.openexchange.usm.api.exceptions.USMStartupException;
import com.openexchange.usm.api.ox.json.OXJSONPropertyNames;
import com.openexchange.usm.api.session.Session;
import com.openexchange.usm.api.session.SessionManager;
import com.openexchange.usm.connector.commands.CommandHandler;
import com.openexchange.usm.json.response.ResponseObject;
import com.openexchange.usm.json.response.ResponseStatusCode;
import com.openexchange.usm.util.AccessLimiter;

public class USMJSONServlet extends HttpServlet {

    private static final long ONE_SECOND = 1000L;

    private static final String COMMAND = "action";

    /**
     *
     */
    private static final long serialVersionUID = 1L;

    private static AtomicReference<ExecutorService> EXECUTOR_SERVICE_REF = new AtomicReference<ExecutorService>();

    /**
     * Sets the specified <code>ExecutorService</code> instance
     *
     * @param executorService The <code>ExecutorService</code> instance
     */
    public static void setExecutorService(final ExecutorService executorService) {
        EXECUTOR_SERVICE_REF.set(executorService);
    }

    /**
     * Gets the <code>ExecutorService</code> instance
     *
     * @return The <code>ExecutorService</code> instance or <code>null</code>
     */
    public static ExecutorService getExecutorService() {
        return EXECUTOR_SERVICE_REF.get();
    }

    private final transient JSONSessionInitializer _sessionInitializer = new JSONSessionInitializer(this);

    private transient SessionManager _sessionManager;

    private transient ContentTypeManager _contentTypeManager;

    private static final transient Log LOG = LogFactory.getLog(USMJSONServlet.class);

    private transient ConfigurationManager _configurationManager;

    private transient long[] _pingTimestamps;

    private transient int _pingTimestampsIndex = -1;

    private final transient Object _pingSynchronizeObject = new Object();

    private transient int _minPingInterval;

    private transient int _maxPingInterval;

    private transient int _maxPingRequestsPerSecond;

    private transient int _pingSampleSize;

    private transient int _maxLimit;

    private transient int _emailSizeLimit;

    private transient int _pimAttachmentCountLimit;

    private transient int _timefilterLimitPast;

    private transient int _timefilterLimitFuture;
    
    private transient int _timefilterCalendarLimitPast;

    private transient int _timefilterCalendarLimitFuture;

    private transient int _inboxPollingInterval;

    private transient int _maxEmailSize;

    private transient int _maxInlineAttachmentSize;

    private transient AccessLimiter _accessLimiter;
    
    /**
     * the activate method; register the httpServer object, set the Servlet alias
     *
     * @param sessionManager
     * @param contentTypeManager
     * @param log
     * @param httpService
     * @param configurationManager
     */
    public USMJSONServlet(SessionManager sessionManager, ContentTypeManager contentTypeManager, ConfigurationManager configurationManager) {
        _sessionManager = sessionManager;
        _contentTypeManager = contentTypeManager;
        _configurationManager = configurationManager;
        try {
            _minPingInterval = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_PING_MIN_INTERVAL,
                ConfigurationProperties.JSON_SERVLET_PING_MIN_INTERVAL_DEFAULT,
                true);
            _maxPingInterval = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_PING_MAX_INTERVAL,
                ConfigurationProperties.JSON_SERVLET_PING_MAX_INTERVAL_DEFAULT,
                true);
            _maxPingRequestsPerSecond = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_PING_MAX_REQUEST_PER_SECOND,
                ConfigurationProperties.JSON_SERVLET_PING_MAX_REQUEST_PER_SECOND_DEFAULT,
                true);
            _pingSampleSize = (int) Math.ceil(ONE_SECOND / (double) _maxPingRequestsPerSecond);
            _pingTimestamps = new long[_pingSampleSize * _maxPingRequestsPerSecond];
            _maxLimit = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_SYNC_MAX_LIMIT_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_SYNC_MAX_LIMIT_DEFAULT,
                false);
            _emailSizeLimit = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_EMAIL_SIZE_LIMIT_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_EMAIL_SIZE_LIMIT_DEFAULT,
                false);
            _pimAttachmentCountLimit = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_PIM_ATTACHMENT_COUNT_LIMIT_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_PIM_ATTACHMENT_COUNT_LIMIT_DEFAULT,
                false);
            _timefilterLimitPast = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_LIMIT_PAST_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_LIMIT_PAST_DEFAULT,
                false);
            _timefilterLimitFuture = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_LIMIT_FUTURE_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_LIMIT_FUTURE_DEFAULT,
                false);
            _timefilterCalendarLimitPast = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_CALENDAR_LIMIT_PAST_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_CALENDAR_LIMIT_PAST_DEFAULT,
                false);
            _timefilterCalendarLimitFuture = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_CALENDAR_LIMIT_FUTURE_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_TIMEFILTER_CALENDAR_LIMIT_FUTURE_DEFAULT,
                false);
            _inboxPollingInterval = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_INBOX_POLLING_INTERVAL_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_INBOX_POLLING_INTERVAL_DEFAULT,
                false);
            _maxEmailSize = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_MAX_EMAIL_SIZE_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_MAX_EMAIL_SIZE_DEFAULT,
                false);
            _maxInlineAttachmentSize = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_MAX_INLINE_ATTACHMENT_SIZE_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_MAX_INLINE_ATTACHMENT_SIZE_DEFAULT,
                false);
            int testInterval = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_TEST_INTERVAL_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_TEST_INTERVAL_DEFAULT,
                false);
            int maxAccesses = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_MAX_ACCESSES_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_MAX_ACCESSES_DEFAULT,
                false);
            int refusalInterval = _configurationManager.getProperty(
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_REFUSAL_INTERVAL_PROPERTY,
                ConfigurationProperties.JSON_SERVLET_ACCESS_CHECK_REFUSAL_INTERVAL_DEFAULT,
                false);
            _accessLimiter = new AccessLimiter(LOG, testInterval, maxAccesses, refusalInterval);
        } catch (USMInvalidConfigurationException e1) {
            logAndThrowStartupError(
                ConnectorBundleErrorCodes.INVALID_ALIAS_FOR_SERVLET,
                e1.getMessage(),
                e1);
        }
    }

    private void logAndThrowStartupError(int errorCode, String detailMessage, Throwable cause) {
        String message = "Open-Xchange USM-Connector couldn't be started: " + detailMessage;
        LOG.error(message, cause);
        deactivate();
        throw new USMStartupException(errorCode, message, cause);
    }

    /**
     * set the servlet alias to null
     */
    public void deactivate() {
        _sessionManager = null;
        _contentTypeManager = null;
        _configurationManager = null;
        // _journal = null; // Do not remove _journal reference to allow logging in other threads to continue as long as possible
        _pingTimestamps = null;
    }

    @Override
    protected void doPut(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        long startTime = System.currentTimeMillis();
        String command = getCommandName(request);
        Session session = null;
        if (LOG.isDebugEnabled()) {
            LOG.debug("HTTP PUT " + command);
        }
        try {
            session = handleCommand(request, response, command);
        } catch (Exception error) {
            LOG.error("Open-Xchange USM/JSON: Internal server error in " + command, error);
            sendError(response, HttpStatus.SC_INTERNAL_SERVER_ERROR);
        }
        
        if (LOG.isDebugEnabled()) {
            LOG.debug(generateTimeLog(session, command, System.currentTimeMillis() - startTime));
        }
    }
    
    private Session handleCommand(HttpServletRequest request, HttpServletResponse response, String command) throws JSONException, IOException {
        ResponseObject result;
        CommandHandler handler = null;
        boolean printedError = false;
        try {
            verifyContentType(request.getContentType());
            handler = CommandDelegateFactory.createCommandHandler(this, command, request, response);
            result = handler.handleRequest();
        } catch (USMJSONAPIException e) {
            result = new ResponseObject(e);
        } catch (USMResourceAccessException e) {
            JSONObject errorDetailsObject = new JSONObject();
            USMException.appendStacktrace(e, errorDetailsObject);
            result = new ResponseObject(
                ResponseStatusCode.INTERNAL_ERROR,
                ConnectorBundleErrorCodes.EXCEPTION_IN_SERVLET_2,
                e.getMessage(),
                errorDetailsObject,
                null);
            LOG.error(computeLogMessage(command, handler, result), e);
            printedError = true;
        } catch (Exception e) {
            JSONObject errorDetailsObject = new JSONObject();
            USMException.appendStacktrace(e, errorDetailsObject);
            result = new ResponseObject(
                ResponseStatusCode.INTERNAL_ERROR,
                ConnectorBundleErrorCodes.EXCEPTION_IN_SERVLET,
                e.getMessage(),
                errorDetailsObject,
                null);
            LOG.error(computeLogMessage(command, handler, result), e);
            printedError = true;
        }
        if (result != null) {
            logResult(command, result, handler, printedError);
            if (handler != null) {
                handler.disposeResources();
            }
            sendResponse(response, result);
        }
        return (handler == null) ? null : handler.getSession();
    }

    private static void sendResponse(HttpServletResponse response, ResponseObject result) throws JSONException, IOException {
        response.setContentType(ServletConstants.USM_JSON_CONTENT_TYPE);
        long contentLength = computeContentLength(result);
        if(contentLength > 0L) {
            if(contentLength <= Integer.MAX_VALUE) {
                response.setContentLength((int) contentLength);
            } else {
                response.setHeader( "Content-Length", String.valueOf(contentLength));
            }
        }
        Writer w = response.getWriter();
        if (w != null) {
            result.write(w);
            w.flush();
            w.close();
        }
    }

    protected void logResult(String command, ResponseObject result, CommandHandler handler, boolean printedError) throws JSONException {
        switch (result.getStatusCode()) {
        case SUCCESS:
        case ACCESS_DENIED:
        case BAD_REQUEST:
        case SYNC_FAILED:
        case UNKNOWN_SESSION:
        case UNKNOWN_SYNCID:
        case UNKNOWN_UUID:
        case WRONG_MISSING_PARAMETERS:
        case UNSUPPORTED_VERSION:
        case SUCCESS_FILTERED_OUT_LARGE_OBJECTS:
            if (LOG.isDebugEnabled()) {
                String message = computeLogMessage(command, handler, result);
                USMJSONException e = result.getCause();
                if (e instanceof Throwable) {
                    LOG.debug(message, (Throwable) e);
                } else {
                    LOG.debug(message);
                }
            }
            break;
        case DATABASE_ERROR:
        case INTERNAL_ERROR:
        case OX_SERVER_ERROR:
        default:
            if (!printedError) {
                String message = computeLogMessage(command, handler, result);
                USMJSONException e = result.getCause();
                if (e instanceof USMJSONAccessDeniedException) {
                    LOG.info(result.getErrorMessage());
                } else if (e instanceof Throwable) {
                    LOG.error(message, (Throwable) e);
                } else {
                    LOG.error(result.toString(3));
                    LOG.error(message);
                }
            }
            break;
        }
    }

    private static long computeContentLength(ResponseObject result) throws JSONException, IOException {
        ByteCountOutputStream counter = new ByteCountOutputStream();
        OutputStreamWriter outputStreamWriter = new OutputStreamWriter(counter, ServletConstants.UTF_ENCODING_8);
        result.write(outputStreamWriter);
        outputStreamWriter.close();
        return counter.getByteCount();
    }

    private static void sendError(HttpServletResponse response, int status) throws IOException {
        if (!response.isCommitted()) {
            try {
                response.sendError(status);
            } catch (IllegalStateException ignored) {
                // may happen if committed by another thread within this methods execution
            }
        }
    }

    private static String computeLogMessage(String command, CommandHandler handler, ResponseObject result) {
        Session session = (handler == null) ? null : handler.getSession();
        StringBuilder sb = new StringBuilder(256);
        sb.append(session).append(' ').append(command).append(' ').append(result.getStatusCode());
        if (result.getStatusCode() != ResponseStatusCode.SUCCESS) {
            sb.append(' ').append(result.getErrorCode());
            if (result.getErrorMessage() != null) {
                sb.append(' ').append(result.getErrorMessage());
            }
        }
        return sb.toString();
    }

    private static String getCommandName(HttpServletRequest request) {
        String[] result = request.getParameterValues(COMMAND);
        if (result != null && result.length == 1) {
            return result[0];
        }
        String command = request.getPathInfo();
        // Remove all leading '/' in the sub-path (not just the first, in case the client uses a wrong configuration)
        int i = 0;
        int length = command.length();
        while (length > i && command.charAt(i) == '/') {
            i++;
        }
        if (i > 0) {
            command = command.substring(i);
        }
        return command;
    }

    protected static void verifyContentType(String contentType) throws USMJSONAPIException {
        if (contentType != null) {
            int semiColon = contentType.indexOf(';');
            if (semiColon >= 0) {
                // Only use the content-type before a ";", ignore any extra information
                contentType = contentType.substring(0, semiColon);
            }
        }
        if (!ServletConstants.TEXT_JAVASCRIPT.equals(contentType)) {
            throw new USMJSONAPIException(
                ConnectorBundleErrorCodes.BAD_REQUEST,
                ResponseStatusCode.BAD_REQUEST,
                "Request type not \"text/javascript\"");
        }
    }

    public SessionManager getSessionManager() throws TemporaryDownOrBusyException {
        SessionManager m = _sessionManager;
        if (m != null)
            return m;
        throw new TemporaryDownOrBusyException(
            ConnectorBundleErrorCodes.SERVLET_SESSIONMANAGER_NOT_AVAILABLE,
            "USM SessionManager not available",
            null);
    }

    public ContentTypeManager getContentTypeManager() {
        return _contentTypeManager;
    }

    public Log getJournal() {
        return LOG;
    }

    public ConfigurationManager getConfigurationManager() {
        return _configurationManager;
    }

    public Session getSession(String user, String password, String device, String clientIP, Map<String, String> xHeaders) throws USMException {
        Session session = getSessionManager().getSession(user, password, PROTOCOL, device, _sessionInitializer, clientIP, xHeaders);
        return session;
    }

    public long getPingInterval() {
        synchronized (_pingSynchronizeObject) {
            int nextIndex = (_pingTimestampsIndex + 1 >= _pingTimestamps.length) ? 0 : _pingTimestampsIndex + 1;
            if (_pingTimestamps[nextIndex] == 0L) {
                return _minPingInterval;
            }
            long deltaT = Math.max(_pingTimestamps[_pingTimestampsIndex] - _pingTimestamps[nextIndex], 1L);
            if (deltaT >= ONE_SECOND * _pingSampleSize) {
                return _minPingInterval;
            }
            return Math.min(_maxPingInterval, _minPingInterval * ONE_SECOND / deltaT);
        }
    }

    public void newPingRequestArrived() {
        synchronized (_pingSynchronizeObject) {
            _pingTimestampsIndex++;
            if (_pingTimestampsIndex >= _pingTimestamps.length) {
                _pingTimestampsIndex = 0;
            }
            _pingTimestamps[_pingTimestampsIndex] = System.currentTimeMillis();
        }
    }

    private static String generateTimeLog(Session session, String command, long duration) {
        if (session == null) {
            return "No session: " + command + " executed in " + duration + " ms.";
        }
        Object numberOfCalls = session.getCustomProperty(OXJSONPropertyNames.NUMBER_OF_CALLS);
        Object accumulatedTime = session.getCustomProperty(OXJSONPropertyNames.ACCUMULATED_TIME);
        if (numberOfCalls == null || accumulatedTime == null) {
            return session + ": " + command + " executed in " + duration + " ms, no OX calls made.";
        }
        return session + ": " + command + " executed in " + duration + " ms, " + numberOfCalls + " OX calls took " + accumulatedTime + " ms.";
    }

    public int getMaxLimit() {
        return _maxLimit;
    }

    public int getEmailSizeLimit() {
        return _emailSizeLimit;
    }

    public int getPimAttachmentCountLimit() {
        return _pimAttachmentCountLimit;
    }

    public int getTimefilterLimitPast() {
        return _timefilterLimitPast;
    }

    public int getTimefilterLimitFuture() {
        return _timefilterLimitFuture;
    }
    
    
    public int getTimefilterCalendarLimitPast() {
        return _timefilterCalendarLimitPast;
    }

    
    public int getTimefilterCalendarLimitFuture() {
        return _timefilterCalendarLimitFuture;
    }

    public int getInboxPollingInterval() {
        return _inboxPollingInterval;
    }

    public AccessLimiter getAccessLimiter() {
        return _accessLimiter;
    }

    public int getMaxEmailSize() {
        return _maxEmailSize;
    }

    public int getMaxInlineAttachmentSize() {
        return _maxInlineAttachmentSize;
    }
}
